#include <gdbm.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define	bool	char			/* boolean type */
#define	not	!			/* logical negation operator */
#define	and	&&			/* logical conjunction */
#define	or	||			/* logical disjunction */
#define	TRUE	(0 == 0)
#define	FALSE	(not TRUE)
#define	loop	while (TRUE)		/* loop until break */
#define	EOS	'\0'			/* end-of-string char */
#ifndef	NULL
#define	NULL	0			/* invalid pointer */
#endif


#define	LINEINC	128


#define	unless(p)	if (not (p))
#define	until(p)	while (not (p))
#define	cardof(a)	(sizeof(a) / sizeof(*(a)))
#define	endof(a)	((a) + cardof(a))
#define	bitsof(a)	(sizeof(a) * 8)


static	char	*me;

static	void	usage(void);
static	char	*mgetline(FILE *fp);
static	void	die(char *fmt, ...);
static	void	warn(char *fmt, ...);
static	void	*ralloc(void *p, uint size);
static	void	*alloc(uint size);

static	bool	parseline(char *lp, char **key, int *bytes);
static	void	storevalue(GDBM_FILE db, char *key, int bytes);


int
main(int argc, char **argv)
{
	GDBM_FILE	db;
	char		*db_path,
			*lp,
			*key;
	int		bytes;

	me = *argv++;
	db_path = *argv;
	if (db_path == NULL)
		usage();
	db = gdbm_open(db_path, 0, GDBM_WRCREAT, 0644, NULL);
	if (db == NULL)
		die("gdbm open failed\n");
	for (;;) {
		lp = mgetline(stdin);
		if (lp == NULL)
			break;
		if (parseline(lp, &key, &bytes)) {
			warn("key is %s, bytes is %d", key, bytes);
			storevalue(db, key, bytes);
			free(key);
		} else
			warn("invalid line: %s", lp);
		free(lp);
	}
	return (0);
}


static void
storevalue(GDBM_FILE db, char *key, int bytes)
{
	datum	dkey,
		dval;

	dkey.dsize = strlen(key) + 1;
	dkey.dptr = key;
	dval.dsize = bytes;
	dval.dptr = alloc(bytes);
	if (gdbm_store(db, dkey, dval, GDBM_INSERT))
		die("gdbm_store of %s failed: %s", key,
		    gdbm_strerror(gdbm_errno));
	free(dval.dptr);
}


static	bool
parseline(char *lp, char **keyp, int *bytes)
{
	char	*tp,		/* pointer to tab */
		*key;

	for (tp = lp; *tp != '\t'; tp++)
		if (*tp == '\0')
			return (FALSE);
	key = alloc(tp-lp);
	strncpy(key, lp, tp-lp);
	key[tp-lp] = '\0';
	*keyp = key;
	*bytes = atoi(tp+1);

	return (TRUE);
}


static	void
usage(void)
{
	die("Usage: %s test_gdbm_file <key_file\n", me);
}



static	char	*
mgetline(FILE *fp)
{
	char	*lbuf,
		*next,
		*limit;
	uint	len;
	int	ch;

	lbuf = (char *)alloc(LINEINC);
	next = lbuf;
	limit = &next[LINEINC];
	do {
		if (next == limit) {
			len = next - lbuf;
			lbuf = (char *)ralloc(lbuf, len + LINEINC);
			next = &lbuf[len];
			limit = &next[LINEINC];
		}
		*next++ = ch = getc(fp);
	} while (ch != '\n' && ch != EOF);
	--next;
	if (ch == EOF  &&  next == lbuf) {
		free((char *)lbuf);
		return (NULL);
	}
	*next++ = EOS;
	lbuf = (char *)ralloc(lbuf, next - lbuf);
	return (lbuf);
}


static	void
warn(char *fmt, ...)
{
	va_list	args;

	fflush(stdout);
	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);
	fprintf(stderr, "\n");
}


static	void
die(char *fmt, ...)
{
	va_list	args;

	fflush(stdout);
	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);
	fprintf(stderr, "\n");
	exit(1);
}


static	void	*
alloc(uint size)
{
	void	*res;

	res = malloc(size);
	if (res == NULL)
		die("Out of space");
	return (res);
}


static	void	*
ralloc(void *p, uint size)
{
	void	*res;

	res = realloc(p, size);
	if (res == NULL)
		die("Out of space");
	return (res);
}
